#include "StdAfx.h"
#include "MAPI.h"

#ifndef RELEASE_OBJ
#define RELEASE_OBJ(s)  \
    if (s != NULL)      \
    {                   \
        s->Release();   \
        s = NULL;       \
    }
#endif //RELEASE_OBJ


using namespace MAPI;

//*****************************************************************************
// [eBeB֐Q
//*****************************************************************************
/**
 * QueryRows
 */
BOOL MAPI::QueryRows(IMAPITable *pTable, LPSRowSet *lppRows)
{
	HRESULT hr = pTable->QueryRows(1, 0, lppRows);
	if (SUCCEEDED(hr)) {
		if ((*lppRows)->cRows != 1) {
			FreeProws(*lppRows);
			return FALSE;
		}
	}
	return SUCCEEDED(hr);
}

/**
 * vector̒gׂĂ[X
 */
template <class T> void MAPI::Release(std::vector<T> *pVector)
{
	if (pVector != NULL) {
		std::vector<T>::iterator it;
		for (it = pVector->begin(); it != pVector->end(); it++) {
			free(*it);
		}

		free(pVector);
		pVector = NULL;
	}
}

//*****************************************************************************
// CMAPI
//*****************************************************************************
CMAPI::CMAPI(void) :
	m_AlreadyInitialized(FALSE)
{
	HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
	if (hr == S_FALSE) {
		//ɏς̂ߏs
		m_AlreadyInitialized = TRUE;
	} else if (FAILED(hr)) {
		//G[
		throw _T("MAPȈɎs܂B");
	}

	MAPIInitialize(NULL);
}

CMAPI::~CMAPI(void)
{
	MAPIUninitialize();

	if (!m_AlreadyInitialized) {
		CoUninitialize();
	}
}

//*****************************************************************************
// CBase
//*****************************************************************************
std::wstring CBase::GetDisplayName(void)
{
	return GetStringProperty(PR_DISPLAY_NAME);
}

BOOL CBase::GetProperty(ULONG ulPropTag, LPSPropValue *lppProp)
{
	LPMAPIPROP lpMapiProp = this->GetMAPIProp();
	HRESULT hr = HrGetOneProp(lpMapiProp, ulPropTag, lppProp);
	return SUCCEEDED(hr);
}

BOOL CBase::SetProperty(LPSPropValue lpProp)
{
	LPMAPIPROP lpMapiProp = this->GetMAPIProp();
	HRESULT hr = HrSetOneProp(lpMapiProp, lpProp);
	return SUCCEEDED(hr);
}

std::wstring CBase::GetStringProperty(ULONG ulPropTag)
{
	LPSPropValue lpProp;
	std::wstring result = _T("");
	if (GetProperty(ulPropTag, &lpProp)) {
		result = lpProp->Value.LPSZ;
		MAPIFreeBuffer(lpProp);
	}
	return result;
}

void CBase::GetSystimeProperty(ULONG ulPropTag, LPSYSTEMTIME lpSysTime)
{
	LPSPropValue lpProp;
	if (GetProperty(ulPropTag, &lpProp)) {
		FileTimeToSystemTime(&lpProp->Value.ft, lpSysTime);
		MAPIFreeBuffer(lpProp);
	}
}

BOOL CBase::GetEntryId(LPSPropValue *lppProp)
{
	return GetProperty(PR_ENTRYID, lppProp);
}

//*****************************************************************************
// CSession
//*****************************************************************************
CSession::CSession(void) :
	m_pSession(NULL),
	m_pMsgStoresTable(NULL),
	m_pAccounts(NULL)
{
	HRESULT hr;

	//OC
	hr = MAPILogonEx(0, NULL, NULL, 0, &m_pSession);
	if (SUCCEEDED(hr)) {
		hr = m_pSession->GetMsgStoresTable(0, &m_pMsgStoresTable);
		if (SUCCEEDED(hr)) {
			SizedSPropTagArray(2, Columns) = {2, {PR_DISPLAY_NAME, PR_ENTRYID}};
			hr = m_pMsgStoresTable->SetColumns((LPSPropTagArray)&Columns, 0);
		}
	}

	if (FAILED(hr)) {
		Release();
	}
}

CSession::~CSession(void)
{
	Release();
}

void CSession::Release(void)
{
	MAPI::Release(m_pAccounts);

	RELEASE_OBJ(m_pMsgStoresTable);
	RELEASE_OBJ(m_pSession);
}

CMsgStores &CSession::GetAccounts(void)
{
	if (m_pAccounts != NULL) {
		return *m_pAccounts;
	}

	m_pAccounts = new CMsgStores();
	LPSRowSet pRows = NULL;
	while (TRUE) {
		if (!QueryRows(m_pMsgStoresTable, &pRows)) break;

		CMsgStore *pMsgStore = new CMsgStore(m_pSession, pRows->aRow[0].lpProps[1].Value.bin);
		m_pAccounts->push_back(pMsgStore);
		FreeProws(pRows);
	}

	return *m_pAccounts;
}

CMsgStore &CSession::GetAccount(const std::wstring &AccountName)
{
	CMsgStores &accounts = GetAccounts();
	CMsgStores::iterator it;
	for (it = accounts.begin(); it != accounts.end(); it++) {
		std::wstring disp = (*it)->GetDisplayName();
		if (AccountName == disp) {
			return **it;
		}
	}

	WCHAR msg[512];
	wsprintf(msg, _T("%s\nLAJEg܂łB"), AccountName.c_str());
	throw msg;
}

CMsgStore &CSession::GetPdxAccount(void)
{
	CMsgStores &accounts = GetAccounts();
	CMsgStores::iterator it;
	for (it = accounts.begin(); it != accounts.end(); it++) {
		if (IsPdxAccountName((*it)->GetDisplayName())) {
			return **it;
		}
	}

	throw _T("PDX̃AJEg܂łB");
}

CFolder &CSession::GetFolder(const std::wstring &FolderPath)
{
	Strings path = SplitPath(FolderPath);
	Strings::iterator it = path.begin();
	if (it == path.end()) {
		WCHAR msg[512];
		wsprintf(msg, _T("%s\nLtH_܂łB"), FolderPath.c_str());
		throw msg;
	}

	CMsgStore &account = GetAccount(*(it++));
	CFolder *pParentFolder, *pSubFolder;
	pParentFolder = &account.GetRootFolder();
	while (it != path.end()) {
		pSubFolder = &(pParentFolder->GetSubFolder(*(it++)));
		pParentFolder = pSubFolder;
	}

	return *pParentFolder;;
}

CFolder &CSession::GetPdxRecievedFolder(void)
{
	CMsgStore &account = GetPdxAccount();
	CFolder &inbox = account.GetInboxFolder();
	return inbox.GetSubFolder(_T("Mς݃ACe"));
}

/**
 * AJEgPDX̂̂ǂׂ
 * @param pAccountName Ώۂ̃AJEg
 * @return PDX̂̂ȂTRUE
 */
BOOL CSession::IsPdxAccountName(const std::wstring &AccountName)
{
	if (AccountName.compare(0, 3, _T("070")) == 0) {
		return TRUE;
	} else {
		return FALSE;
	}
}

CSession::Strings CSession::SplitPath(const std::wstring &FolderPath)
{
	Strings result;
	std::wstring::size_type pos = 0;
	std::wstring buff = FolderPath;

	while ((pos = buff.find(_T("\\"))) != std::wstring::npos) {
		result.push_back(buff.substr(0, pos));
		buff = buff.substr(pos + 1);
	}
	if (buff.size() > 0) {
		result.push_back(buff);
	}

	return result;
}

//*****************************************************************************
// CMsgStore
//*****************************************************************************
CMsgStore::CMsgStore(IMAPISession *pSession, SBinary entry) :
	m_pSession(pSession),
	m_pMsgStore(NULL),
	m_pRootFolder(NULL),
	m_pInboxFolder(NULL)
{
	HRESULT hr = m_pSession->OpenMsgStore(0,
		entry.cb,
		(LPENTRYID) entry.lpb,
		NULL,
		NULL,
		&m_pMsgStore);

	if (FAILED(hr)) {
		Release();
	}
}

CMsgStore::~CMsgStore(void)
{
	Release();
}

void CMsgStore::Release(void)
{
	free(m_pInboxFolder);
	m_pInboxFolder = NULL;
	free(m_pRootFolder);
	m_pRootFolder = NULL;

	RELEASE_OBJ(m_pMsgStore);
	m_pSession = NULL;
}

CFolder &CMsgStore::GetRootFolder(void)
{
	return GetFolder(PR_IPM_SUBTREE_ENTRYID, &m_pRootFolder);
}

CFolder &CMsgStore::GetInboxFolder(void)
{
	return GetFolder(PR_CE_IPM_INBOX_ENTRYID, &m_pInboxFolder);
}

CFolder &CMsgStore::GetFolder(ULONG ulPropTag, CFolder **ppFolder)
{
	if (*ppFolder != NULL) {
		return **ppFolder;
	}

	LPSPropValue lpProps = NULL;
	if (GetProperty(ulPropTag, &lpProps)) {
		*ppFolder = new CFolder(m_pSession, m_pMsgStore, lpProps->Value.bin, GetDisplayName());
		MAPIFreeBuffer(lpProps);
	}

	return **ppFolder;
}

//*****************************************************************************
// CFolder
//*****************************************************************************
CFolder::CFolder(IMAPISession *pSession, LPMAPIFOLDER pFolder, SBinary entry, const std::wstring ParentPath) :
	m_pSession(pSession),
	m_pFolder(NULL),
	m_pHierarchy(NULL),
	m_pContents(NULL),
	m_pSubFolders(NULL),
	m_Messages(),
	m_Path(ParentPath)
{
	HRESULT hr = pFolder->OpenEntry(
		entry.cb,
		(LPENTRYID) entry.lpb,
		NULL,
		MAPI_BEST_ACCESS,
		NULL,
		(LPUNKNOWN *)&m_pFolder);
	if (SUCCEEDED(hr)) {
		m_Path += _T("\\") + GetDisplayName();
	} else {
		Release();
	}
}

CFolder::CFolder(IMAPISession *pSession, LPMDB pMsgStore, SBinary entry, const std::wstring ParentPath) :
	m_pSession(pSession),
	m_pFolder(NULL),
	m_pHierarchy(NULL),
	m_pContents(NULL),
	m_pSubFolders(NULL),
	m_Messages(),
	m_Path(ParentPath)
{
	HRESULT hr = pMsgStore->OpenEntry(
		entry.cb,
		(LPENTRYID) entry.lpb,
		NULL,
		MAPI_BEST_ACCESS,
		NULL,
		(LPUNKNOWN *)&m_pFolder);
	if (SUCCEEDED(hr)) {
		std::wstring disp = GetDisplayName();
		if (disp != _T("IPM.Root")) {
			m_Path += _T("\\") + GetDisplayName();
		}
	} else {
		Release();
	}
}

CFolder::~CFolder(void)
{
	Release();
}

void CFolder::Release(void)
{
	m_Messages.clear();
	MAPI::Release(m_pSubFolders);

	RELEASE_OBJ(m_pHierarchy);
	RELEASE_OBJ(m_pContents);
	RELEASE_OBJ(m_pFolder);
	m_pSession = NULL;
}

CFolders &CFolder::GetSubFolders(BOOL IsRefresh)
{
	LPSRowSet pRows = NULL;

	if (!IsRefresh && m_pSubFolders != NULL && m_pSubFolders->size() > 0) {
		return *m_pSubFolders;
	}

	MAPI::Release(m_pSubFolders);
	m_pSubFolders = new CFolders();
	IMAPITable *pHierarchy = GetHierarchy();
	while (pHierarchy != NULL) {
		if (!QueryRows(pHierarchy, &pRows)) break;

		CFolder *pFolder = new CFolder(m_pSession, m_pFolder, pRows->aRow[0].lpProps[0].Value.bin, GetPath());
		m_pSubFolders->push_back(pFolder);

		FreeProws(pRows);
	}

	return *m_pSubFolders;
}

CFolder &CFolder::GetSubFolder(const std::wstring &FolderName, BOOL IsRefresh)
{
	CFolders::iterator it;
	CFolders &folders = GetSubFolders(IsRefresh);
	for (it = folders.begin(); it != folders.end(); it++) {
		std::wstring folderName = (*it)->GetDisplayName();
		if (FolderName.compare(folderName) == 0) {
			return **it;
		}
	}

	WCHAR msg[512];
	wsprintf(msg, _T("%s\nLtH_܂łB"), FolderName.c_str());
	throw msg;
}

CMessages &CFolder::GetAllMessages(void)
{
	SizedSPropTagArray(1, tags) = {1, {PR_ENTRYID}};
	m_Messages.clear();
	EnumMessages(EnumMsgAllFunc, (LPSPropTagArray) &tags, &m_Messages);
	return m_Messages;
}

CMessages &CFolder::GetNoreadMessages(void)
{
	SizedSPropTagArray(2, tags) = {2, {PR_ENTRYID, PR_MESSAGE_FLAGS}};
	m_Messages.clear();
	EnumMessages(EnumMsgNoreadFunc, (LPSPropTagArray) &tags, &m_Messages);
	return m_Messages;
}

ULONG CFolder::GetAllMessageCount(void)
{
	HRESULT hr = S_OK;
	IMAPITable *pContents = GetContents();
	ULONG count = 0;
	hr = pContents->GetRowCount(0, &count);
	if (FAILED(hr)) throw _T("bZ[W擾ł܂łB");

	return count;
}

void CFolder::EnumMessages(BOOL (*pEnumMsgFunc)(CFolder &, LPSRow, LPVOID), LPSPropTagArray lpTags, LPVOID lpParam)
{
	HRESULT hr = S_OK;
	LPSRowSet pRows = NULL;

	IMAPITable *pContents = GetContents();

	hr = pContents->SetColumns(lpTags, 0);
	if (FAILED(hr)) throw _T("bZ[W擾Ɏs܂B");
	hr = pContents->SeekRow(BOOKMARK_BEGINNING, 0, NULL);
	if (FAILED(hr)) throw _T("bZ[W擾Ɏs܂B");
	while (pContents != NULL) {
		if (!QueryRows(pContents, &pRows)) break;

		if (!(*pEnumMsgFunc)(*this, &(pRows->aRow[0]), lpParam)) {
			FreeProws(pRows);
			break;
		}

		FreeProws(pRows);
	}
}

BOOL CFolder::CopyMessage(SBinary entry, CFolder &destFolder, ULONG ulFlags)
{
	ENTRYLIST entryList;
	entryList.cValues = 1;
	entryList.lpbin = &entry;

	HRESULT hr = m_pFolder->CopyMessages(&entryList, NULL, destFolder.m_pFolder, 0, NULL, ulFlags);
	return SUCCEEDED(hr);
}

BOOL CFolder::MailSwitchToFolder(void)
{
	LPSPropValue lpProp;
	BOOL result = FALSE;
	if (GetEntryId(&lpProp)) {
		result = SUCCEEDED(::MailSwitchToFolder((LPENTRYID) lpProp->Value.bin.lpb, lpProp->Value.bin.cb));
		MAPIFreeBuffer(lpProp);
	}

	return result;
}

/**
 * TutH_MAPITable擾
 */
IMAPITable* CFolder::GetHierarchy(void)
{
	RELEASE_OBJ(m_pHierarchy);

	HRESULT hr = m_pFolder->GetHierarchyTable(0, &m_pHierarchy);
	if (FAILED(hr)) return NULL;

	static SizedSPropTagArray(1, Columns) = {1, {PR_ENTRYID}};
	hr = m_pHierarchy->SetColumns((LPSPropTagArray)&Columns, 0);
	if (FAILED(hr)) {
		RELEASE_OBJ(m_pHierarchy);
	}

	return m_pHierarchy;
}

/**
 * bZ[WMAPITable擾
 */
IMAPITable* CFolder::GetContents(void)
{
	RELEASE_OBJ(m_pContents);

	HRESULT hr = m_pFolder->GetContentsTable(0, &m_pContents);
	if (FAILED(hr)) return NULL;

	return m_pContents;
}

BOOL CFolder::EnumMsgAllFunc(CFolder &thisFolder, LPSRow pRow, LPVOID lpParam)
{
	CMessage *pMessage = new CMessage(thisFolder.m_pSession, pRow->lpProps[0].Value.bin);
	((CMessages *)lpParam)->push_back(pMessage);
	return TRUE;
}

BOOL CFolder::EnumMsgNoreadFunc(CFolder &thisFolder, LPSRow pRow, LPVOID lpParam)
{
	if (!(pRow->lpProps[1].Value.ul & MSGFLAG_READ)) {
		CMessage *pMessage = new CMessage(thisFolder.m_pSession, pRow->lpProps[0].Value.bin);
		((CMessages *)lpParam)->push_back(pMessage);
	}
	return TRUE;
}



//*****************************************************************************
// CMessage
//*****************************************************************************
CMessage::CMessage(IMAPISession *pSession, SBinary entry) :
	m_pSession(pSession),
	m_pMessage(NULL)
{
	HRESULT hr = m_pSession->OpenEntry(
		entry.cb,
		(LPENTRYID) entry.lpb,
		NULL,
		MAPI_BEST_ACCESS,
		NULL,
		(LPUNKNOWN *)&m_pMessage);
	if (FAILED(hr)) {
		Release();
	}
}

CMessage::~CMessage(void)
{
	Release();
}

void CMessage::Release(void)
{
	RELEASE_OBJ(m_pMessage);
	m_pSession = NULL;
}

std::wstring CMessage::GetSubject(void)
{
	return GetStringProperty(PR_SUBJECT);
}

std::wstring CMessage::GetSenderName(void)
{
	return GetStringProperty(PR_SENDER_NAME);
}

std::wstring CMessage::GetSenderEmailAddress(void)
{
	return GetStringProperty(PR_SENDER_EMAIL_ADDRESS);
}

std::wstring CMessage::GetMessageClass(void)
{
	return GetStringProperty(PR_MESSAGE_CLASS);
}

void CMessage::GetDeliveryTime(LPSYSTEMTIME lpSysTime)
{
	GetSystimeProperty(PR_MESSAGE_DELIVERY_TIME, lpSysTime);
}

BOOL CMessage::IsRead(void)
{
	BOOL result = FALSE;

	LPSPropValue lpProp;
	if (GetProperty(PR_MESSAGE_FLAGS, &lpProp)) {
		result = (lpProp->Value.ul & MSGFLAG_READ);
		MAPIFreeBuffer(lpProp);
	}
	return result;
}

BOOL CMessage::SetRead(BOOL flag)
{
	BOOL result = FALSE;

	LPSPropValue lpProp;
	if (GetProperty(PR_MESSAGE_FLAGS, &lpProp)) {
		if (flag) {
			lpProp->Value.ul |= MSGFLAG_READ;
		} else {
			lpProp->Value.ul &= !MSGFLAG_READ;
		}
		SetProperty(lpProp);
		MAPIFreeBuffer(lpProp);
	}
	return result;
}	

BOOL CMessage::Display(void)
{
	LPSPropValue lpProp;
	BOOL result = FALSE;
	if (GetProperty(PR_ENTRYID, &lpProp)) {
		result = SUCCEEDED(MailDisplayMessage((LPENTRYID) lpProp->Value.bin.lpb, lpProp->Value.bin.cb));
		MAPIFreeBuffer(lpProp);
	}

	return result;
}
